/** @file camogm_jpeg.c
 * @brief Provides writing to series of individual JPEG files for @e camogm
 * @copyright Copyright (C) 2016 Elphel, Inc.
 *
 * @par <b>License</b>
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/** @brief This define is needed to use lseek64 and should be set before includes */
#define THREADS_MODE 2 // 1 // 0 - how it was befor 07.01.2023
#define _LARGEFILE64_SOURCE
//#define __USE_GNU // for O_DIRECT
#define _GNU_SOURCE // for O_DIRECT
#include <linux/raw.h> // for O_DIRECT
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
//#include <bits/fcntl.h> // 'O_DIRECT'
#include <sys/uio.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <string.h>
#include <dirent.h>
#include <time.h>
#include <elphel/c313a.h>
#include <elphel/ahci_cmd.h>

#include "camogm_jpeg.h"
#include "camogm_read.h"
#include "camogm_align.h"
//#define BUFFER_KVEC
// just for Eclipse to be happy
#ifndef O_DIRECT
#define O_DIRECT __O_DIRECT
#endif

#define MAX_IMAGE_ALIGNED (4096*256)
/** State file record format. It includes device path in /dev, starting, current and ending LBAs */
#define STATE_FILE_FORMAT         "%s\t%llu\t%llu\t%llu\n"
/** Status file update period, in seconds */
#define STAT_UPDATE_PERIOD        60

/* forward declarations */
static void *jpeg_writer(void *thread_args);
static int save_state_file(const camogm_state *state);

/**
 * Get next segment page to write to disk - should run under locked mutex
 * @param wparams
 * @return next page or -1 if none are ready
 */
inline int get_write_page(struct writer_params *wparams){
	int i, wpage = wparams->chunk_page_prep;
	for (i = 0; i < SEGMENTS_PAGES; i++){
		if (wparams->chunk_page_state[wpage] == SEGPAGE_FULL) {
			return wpage;
		}
		wpage++;
		if (wpage >= SEGMENTS_PAGES){
			wpage -= SEGMENTS_PAGES;
		}
	}
	return -1;
}


/**
 * Get number of emty pages for the main thread to fill
 * @param state
 * @return number of next empty pages
 */
inline int get_num_empty(struct writer_params *wparams){
	int i, wpage = wparams->chunk_page_prep;
	for (i = 0; i < SEGMENTS_PAGES; i++){
		if (wparams->chunk_page_state[wpage] != SEGPAGE_EMPTY) {
			return i;
		}
		wpage++;
		if (wpage >= SEGMENTS_PAGES){
			wpage -= SEGMENTS_PAGES;
		}
	}
	return SEGMENTS_PAGES;
}
inline int get_num_busy(struct writer_params *wparams){
	int i, nbusy = 0;
	for (i = 0; i < SEGMENTS_PAGES; i++){
		if (wparams->chunk_page_state[i] == SEGPAGE_BUSY) {
			nbusy++;
		}
	}
	return nbusy;
}



/**
 * Emulate writev() by consecutive write()
 * @param __fd
 * @param __iovec
 * @param __count
 * @return
 */
ssize_t emul_writev (int fd, const struct iovec *iovec, int count) {
	int i;
	ssize_t wlen, tot_len = 0;
	for (i = 0; i < count; i++) {
		if (iovec[i].iov_len > 0) {
			if (iovec[i].iov_len & (PAGE_PHYS -1)){
				D0(fprintf(debug_file, "Length not page-aligned: while trying direct writing of %d bytes from %p, data = 0x%llx, segment=%d\n",\
						iovec[i].iov_len, iovec[i].iov_base, ((uint64_t*) iovec[i].iov_base)[0],i ));
				D0(syslog (LOG_INFO, "Length not page-aligned: while trying direct writing of %d bytes from %p, data = 0x%llx, segment=%d\n",\
						iovec[i].iov_len, iovec[i].iov_base, ((uint64_t*) iovec[i].iov_base)[0],i));
				debug_level = 6;
				return -1;
			}
			wlen = write (fd, iovec[i].iov_base, iovec[i].iov_len);
			if (wlen < 0) {
				D0(fprintf(debug_file, "Error %d while trying direct writing of %d bytes from %p, data = 0x%llx, segment=%d\n",\
						wlen, iovec[i].iov_len, iovec[i].iov_base, ((uint64_t*) iovec[i].iov_base)[0],i ));
				return wlen;
			} else if (wlen < iovec[i].iov_len) {
				return tot_len + wlen;
			} else {
				tot_len += wlen;
			}
		}
	}
	return tot_len;
}

/** Get starting and endign LBAs of the partition specified as raw device buffer */
static int get_disk_range(struct range *range)
{
	FILE *f;

	// get raw device buffer starting postion on disk
	f = fopen(SYSFS_AHCI_LBA_START, "r");
	if (f == NULL) {
		return -1;
	}
	fscanf(f, "%llu\n", &range->from);
	fclose(f);

	// get raw device buffer ending postion on disk
	f = fopen(SYSFS_AHCI_LBA_END, "r");
	if (f == NULL) {
		return -1;
	}
	fscanf(f, "%llu\n", &range->to);
	fclose(f);

	return 0;
}

/** Get write pointer from a file. This functions check not only the name of a partition, but
 * its geometry as well */
static int find_state(FILE *f, uint64_t *pos, const rawdev_buffer *rawdev)
{
	size_t len;
	uint64_t start_pos, curr_pos, end_pos;
	struct range range;
	char buff[ELPHEL_PATH_MAX];
	char dev_name[ELPHEL_PATH_MAX];

	if (f == NULL || pos == NULL)
		return -1;
	if (get_disk_range(&range) != 0) {
		return -2;
	}
	D3(fprintf(debug_file, "get_disk_range() -> from = %llu, to=%llu\n", range.from, range.to));
	// skip first line containing file header
	fgets(buff, ELPHEL_PATH_MAX, f);
	D3(fprintf(debug_file, "find_state(): Got first string '%s' from state file %s\n", buff,  rawdev->state_path));
	while (fgets(buff, ELPHEL_PATH_MAX, f) != NULL) {
		D3(fprintf(debug_file, "find_state(): Got next string '%s' from state file %s\n", buff,  rawdev->state_path));
		sscanf(buff, STATE_FILE_FORMAT, dev_name, &start_pos, &curr_pos, &end_pos);
		len = strlen(dev_name);
		if (strncmp(rawdev->rawdev_path, dev_name, len) == 0 &&
				range.from == start_pos &&
				range.to == end_pos) {
			D3(fprintf(debug_file, "find_state():got match, curr_pos= %llu from=%llu, to=%llu\n", curr_pos, range.from, range.to));
			*pos = curr_pos;
			return 0;
		} else {
			D3(fprintf(debug_file, "find_state():no match, dev_name=%s(%s), from=%llu(%llu), to=%llu(%llu) curr_pos= %llu\n",\
					dev_name, rawdev->rawdev_path, start_pos, range.from, end_pos, range.to, curr_pos));
		}
	}
	D3(fprintf(debug_file, "find_state():no match, return -2"));
	return -3;
}

int get_dir_path(char* file_path, char* directory){ // directory should be allocated, returns w/o trailing '/'
//	int dirlen;
	char * cp;
	strncpy(directory, file_path, ELPHEL_PATH_MAX); // rawdev->state_path is always absolute
	cp=strrchr(directory,'/');
	if (!cp) {
		return -1; // no '/'
	}
	*cp=0; // truncate at '_'
	return strlen(directory);
}

int getup_state_files(camogm_state *state) { //rawdev_buffer *rawdev){
	rawdev_buffer *rawdev = &state->rawdev;
	int dirlen, preflen; // length of directory path including trailing '/'
	char * cp;
	char pathname[ELPHEL_PATH_MAX];
	char last_path[ELPHEL_PATH_MAX];
	char state_directory[ELPHEL_PATH_MAX];
	DIR *d;
	struct dirent *dir;
	struct stat fstat;
	__time_t tsec = 0;
#if 0
	strncpy(state_directory, rawdev->state_path, ELPHEL_PATH_MAX); // rawdev->state_path is always absolute
	cp=strrchr(state_directory,'/');
	if (!cp) {
		return -1;
	}
	*cp=0; // truncate
	dirlen = strlen(state_directory);
#endif
	dirlen = get_dir_path(rawdev->state_path, state_directory);
	preflen = dirlen+1; // counts '/'
	strcpy(pathname,state_directory);
	cp = pathname + dirlen;
	*cp = '/';
	cp++;
	*cp=0;
	//	return rawdev->state_directory;
	// get latest
	strcpy(rawdev->state_prefix, state_directory);
	rawdev->state_index = 0;
	cp = rawdev->state_prefix + strlen(rawdev->state_prefix); // right after '/'
	*cp++='/';
	strncpy(cp, STATE_PREF_DFLT,ELPHEL_PATH_MAX - dirlen - strlen(STATE_PREF_DFLT) - 1);
	D3(syslog (LOG_INFO, "_s00_ state_path='%s' state_directory='%s'", \
			rawdev->state_path, state_directory));
	d = opendir(state_directory);
	if (d) {
		while ((dir = readdir(d)) != NULL) {
//			printf("%s\n", dir->d_name);
			// STATE_EXT
			if (!strcmp(STATE_EXT, dir->d_name +(strlen(dir->d_name) - strlen(STATE_EXT)))) { // same extension
				strncpy(pathname+preflen, dir->d_name, ELPHEL_PATH_MAX-preflen-strlen(dir->d_name) - 1);
				if (strcmp(pathname, rawdev->state_path)) { // skip camogm.disk itself
					if (!stat(pathname, &fstat)) {
						if (S_ISREG(fstat.st_mode) && (fstat.st_mtim.tv_sec > tsec)) {
							tsec = fstat.st_mtim.tv_sec;
							strcpy (last_path,pathname);
						}
						D3(syslog (LOG_INFO,"_s01_ pathname='%s' last_path='%s',fstat.st_mtim.tv_sec=%ld,  tsec=%ld", \
								pathname, last_path, fstat.st_mtim.tv_sec, tsec));
					}
				}
			}
		}
		closedir(d);
	}
	if (tsec > 0 ) { // at least one exists)
		cp = strrchr(last_path,'_'); // find last '_'
		if (!cp || strchr(cp,'/')){ //'_' in the directory name
			cp = last_path+strlen(last_path)-strlen(STATE_EXT); // truncate at '.'
		} else {
			sscanf(cp+1, "%d",&rawdev->state_index);
		}
		*cp = '\0';
		strcpy(rawdev->state_prefix, last_path);
		rawdev->state_index++;
	}
	//state->path
	preflen = sprintf(state->path,STATE_FMT,rawdev->state_prefix,rawdev->state_index);
	D2(syslog (LOG_INFO,"_s02_ prefix='%s',index=%d, preflen=%d, state->path=%s", \
			rawdev->state_prefix, rawdev->state_index, preflen, state->path));
	return 0;
}

// need to modify - if prefix already appeared, and provided index <= old, increment latest old one
int set_state_file_from_prefix(camogm_state *state, const char * prefix){ // prefix is relative to /dev/sda1 (where camogm.disk is)
	rawdev_buffer *rawdev = &state->rawdev;
	char last_path[ELPHEL_PATH_MAX];
	char * cp, *cp1, *local_pref; // pointer to last_path after '/'
	int preflen, len, i , new_index, old_index;
	DIR *d;
	struct dirent *dir;
//	char pathname[ELPHEL_PATH_MAX];
	char state_directory[ELPHEL_PATH_MAX];
	cp = last_path;
	if (prefix[0] != '/') { // relative
		len = get_dir_path(rawdev->state_path, last_path);
		if (len < 0) {
			D0(syslog (LOG_ERR,"%s:%d: State file '%s' is not absolute, should start with /dev/sda1", __FILE__, __LINE__, rawdev->state_path));
		}
		cp = last_path + len;
		*cp++ = '/';
		strncpy(cp, prefix, ELPHEL_PATH_MAX - len -2);
	} else {
		strcpy(last_path,prefix); // may be either relative to camogm.disk directory or absolute

	}
	// now - always absolute
	if (!strcmp(last_path + (strlen(last_path)-strlen(STATE_EXT)),STATE_EXT)) { // ends with .disk
		last_path[strlen(last_path)-strlen(STATE_EXT)] = '\0'; // truncate at .
	}
    // absolute, w/o '.disk.
	strcpy(rawdev->state_prefix, last_path); // use anyway, only index may be adjusted
//	local_pref = last_path+
	new_index = 0;
	cp = strrchr(last_path,'_'); // find last '_'
	if (!cp || strchr(cp,'/')){ //'_' in the directory name
		cp = last_path+strlen(last_path); // full string w/o optional extension
	} else {
		cp1 = cp+1;
		len = strlen(cp1);
		for (i = 0; i < len; i++) {
			if ((cp1[i] < '0') || (cp1[i] > '9')) {
				cp = NULL;
				break;
			}
		}
		D2(syslog (LOG_INFO,"_s030_ len=%d, cp=0x%08x, cp1='%s'", \
				len, (int) cp, cp1));
		if (cp) {
			sscanf(cp+1, "%d",&new_index);
		} else {
			cp = last_path+strlen(last_path);
		}
	}
	*cp = '\0'; // instead of the '_' or '.' in '.disk'
	// last_path - prefix (absolute), new_index
	// get directory from the new potential prefix (last_path)
	len = get_dir_path(last_path, state_directory);
	if (len < 0) {
		D0(syslog (LOG_ERR,"%s:%d: Prefix '%s' should be absolute here", __FILE__, __LINE__, last_path));
		return -1;
	}
	local_pref = last_path + (len+1); // right after '/', ends at former "_" or ".disk"
	preflen = strlen(local_pref);
	// see if the same prefix already appeared, if yes - find the maximal index
	old_index = -1;
//strcpy(rawdev->state_prefix, state_directory)
	D3(syslog (LOG_INFO, "_s31_ new prefix='%s' new directory='%s'", \
			last_path, state_directory));
	d = opendir(state_directory);
	if (d) {
		while ((dir = readdir(d)) != NULL) {
			if (!strcmp(STATE_EXT, dir->d_name +(strlen(dir->d_name) - strlen(STATE_EXT)))) { // same extension
				if (!strncmp(local_pref, dir->d_name, preflen)) { // starts same
					if ((dir->d_name[preflen] == '_') || (dir->d_name[preflen] == '.')){
						i=0; // if not get any - will be 0;
						sscanf(dir->d_name+preflen+1, "%d",&i);
						D3(syslog (LOG_INFO, "_s32_ matching old file='%s' index=%d", \
								dir->d_name, i));
						if (i > old_index) {
							old_index = i;
						}
					}
				}
			}
		}
		closedir(d);
	}
	if (old_index >= new_index){
		rawdev->state_index = old_index + 1;
		D3(syslog (LOG_INFO, "_s33_ using incremented old index=%d (new was %d), prefix = '%s'", \
				rawdev->state_index, new_index, rawdev->state_prefix));
	} else {
		rawdev->state_index = new_index;
		D3(syslog (LOG_INFO, "_s33_ using new index index=%d (old was %d), prefix = '%s'", \
				rawdev->state_index, old_index, rawdev->state_prefix));
	}
	preflen = snprintf(state->path, ELPHEL_PATH_MAX, STATE_FMT, rawdev->state_prefix ,rawdev->state_index);
	D2(syslog (LOG_INFO,"_s03_ prefix='%s', rawdev->state_prefix='%s', index=%d, preflen=%d, state->path=%s", \
			prefix, rawdev->state_prefix, rawdev->state_index, preflen, state->path));
	return 0;
}

char * get_state_path(camogm_state *state){
//	int preflen;
	rawdev_buffer *rawdev = &state->rawdev;
//	preflen =
	snprintf(state->path, ELPHEL_PATH_MAX, STATE_FMT, rawdev->state_prefix, rawdev->state_index);
	return state->path;
}

void next_state_path(camogm_state *state){
//	int preflen;
	rawdev_buffer *rawdev = &state->rawdev;
	rawdev->state_index++;
//	preflen =
//	snprintf(state->path, ELPHEL_PATH_MAX, STATE_FMT, rawdev->state_prefix, rawdev->state_index);
//	return state->path;
}

int copy_state_file(camogm_state *state){
//	rawdev->state_path-> get_state_path(state); increment - separately
	rawdev_buffer *rawdev = &state->rawdev;
	D2(syslog (LOG_INFO,"_s05_ Copying state files: '%s' -> %s", \
			rawdev->state_path, get_state_path(state)));
	return copy_file(rawdev->state_path, get_state_path(state));
}

int copy_file(const char *from, const char *to){
    int fd_to, fd_from;
    char buf[4096];
    ssize_t nread;
    int saved_errno;

    fd_from = open(from, O_RDONLY);
    if (fd_from < 0)
        return -1;

    fd_to = open(to, O_WRONLY | O_CREAT | O_EXCL, 0666);
    if (fd_to < 0)
        goto out_error;

    while ((nread = read(fd_from, buf, sizeof buf)), nread > 0)
    {
        char *out_ptr = buf;
        ssize_t nwritten;

        do {
            nwritten = write(fd_to, out_ptr, nread);

            if (nwritten >= 0)
            {
                nread -= nwritten;
                out_ptr += nwritten;
            }
            else if (errno != EINTR)
            {
                goto out_error;
            }
        } while (nread > 0);
    }

    if (nread == 0)
    {
        if (close(fd_to) < 0)
        {
            fd_to = -1;
            goto out_error;
        }
        close(fd_from);

        /* Success! */
        return 0;
    }

  out_error:
    saved_errno = errno;

    close(fd_from);
    if (fd_to >= 0)
        close(fd_to);

    errno = saved_errno;
    return -1;
}


/** Read state from file and restore disk write pointer */
int open_state_file(const rawdev_buffer *rawdev, uint64_t *current_lba)
{
	FILE *f;
	int ret = 0;
	uint64_t lba_pos;

	if (strlen(rawdev->state_path) == 0) {
		return ret;
	}

	f = fopen(rawdev->state_path, "r");
	if (f != NULL) {
		if ((ret = find_state(f, &lba_pos, rawdev)) == 0) {
			*current_lba = lba_pos;
			D1(fprintf(debug_file, "Got starting LBA from state file %s: %llu\n", rawdev->state_path, lba_pos));
		}
		fclose(f);
	} else {
		ret = -4;
	}
	return ret;
}

/** Save current position of the disk write pointer */
static int save_state_file(const camogm_state *state)
{
	int ret = 0;
	FILE *f;
	const rawdev_buffer *rawdev = &state->rawdev;
	const struct writer_params *params = &state->writer_params;
	uint64_t current_lba = get_lba_next(params);
	D3(fprintf(debug_file, "save_state_file(): current_lba=%lld lba_start=%lld, lba_end=%lld @ %07d\n", \
			current_lba , params->lba_start, params->lba_end, get_fpga_usec(state->fd_fparmsall[0], 0)));
	if ((current_lba < params->lba_start) || (current_lba > params->lba_end)){
		D0(fprintf(debug_file, "save_state_file(): ERROR ! invalid current_lba=%lld, params->next_segment_pos=%lld @ %07d\n",\
				current_lba, params->next_segment_pos, get_fpga_usec(state->fd_fparmsall[0], 0)));
	}
	D3(syslog (LOG_INFO, "save_state_file(): lba_start = %lld, lba_end=%lld, current_lba=%lld, params->next_segment_pos=%lld",\
			 params->lba_start,  params->lba_end,  get_lba_next(params), params->next_segment_pos));
	if (strlen(rawdev->state_path) == 0) {
		return ret;
	}

	// save pointers to a regular file
	f = fopen(rawdev->state_path, "w");
	if (f == NULL) {
		return -1;
	}
	fprintf(f, "Device\t\tStart LBA\tCurrent LBA\tEnd LBA\n");
//	fprintf(f, STATE_FILE_FORMAT, rawdev->rawdev_path, params->lba_start, params->lba_current, params->lba_end);
	fprintf(f, STATE_FILE_FORMAT, rawdev->rawdev_path, params->lba_start, get_lba_next(params), params->lba_end);
	fflush(f);
	fsync(fileno(f));
	fclose(f);
	D6(fprintf(debug_file,"Device\t\tStart LBA\tCurrent LBA\tEnd LBA\n");\
	fprintf(f, STATE_FILE_FORMAT, rawdev->rawdev_path, params->lba_start, get_lba_next(params), params->lba_end));

	return ret;
}




/**
 * @brief Initialize synchronization resources for disk writing thread and then start this thread. This function
 * is call each time JPEG format is set or changed, thus we need to check the state of writing thread before
 * initialization to prevent spawning multiple threads.
 * @param[in]   state   a pointer to a structure containing current state
 * @return      0 if initialization was successful and negative value otherwise
 */
int camogm_init_jpeg(camogm_state *state)
{
	int ret = 0;
	int ret_val;
	struct writer_params *params = &state->writer_params;
	if (params->state == STATE_STOPPED) {
		ret_val = pthread_cond_init(&params->main_cond, NULL);
		if (ret_val != 0) {
			D0(fprintf(debug_file, "Can not initialize conditional variable for main thread: %s\n", strerror(ret_val)));
			ret = -1;
		}
		ret_val = pthread_cond_init(&params->writer_cond, NULL);
		if (ret_val != 0) {
			D0(fprintf(debug_file, "Can not initialize conditional variable for writing thread: %s\n", strerror(ret_val)));
			ret = -1;
		}
		ret_val = pthread_mutex_init(&params->writer_mutex, NULL);
		if (ret_val != 0) {
			D0(fprintf(debug_file, "Can not initialize mutex for writing thread: %s\n", strerror(ret_val)));
			ret = -1;
		}
//		ret_val = pthread_create(&params->writer_thread, NULL, jpeg_writer, (void *)state);
		if (ret_val != 0) {
			D0(fprintf(debug_file, "Can not start writer thread: %s\n", strerror(ret_val)));
			ret = -1;
		}
		ret_val = init_align_buffers(state);
		if (ret_val != 0) {
			D0(fprintf(debug_file, "Can not initialize alignment buffers\n"));
			ret = -1;
			return ret;
		}
	}

	return ret;
}

/**
 * @brief Stop disk writing thread and free its resources. This function is called after the program receives 'exit' command.
 * @param[in]   state   a pointer to a structure containing current state
 * @return      None
 */
void camogm_free_jpeg(camogm_state *state)
{
	struct writer_params *params = &state->writer_params;
	pthread_cond_destroy(&params->main_cond);
	pthread_cond_destroy(&params->writer_cond);
	pthread_mutex_destroy(&params->writer_mutex);

	// terminate writing threads
#if 0
	pthread_mutex_lock(&params->writer_mutex);
	params->exit_write_threads = true;
	pthread_cond_broadcast(&params->writer_cond);
	pthread_mutex_unlock(&params->writer_mutex);
	for (int i = 0; i < NUM_WRITER_THREADS; i++){
		if (params->writer_threads[i]) {
			pthread_join(params->writer_threads[i], NULL);
		}
	}
#endif
	reset_segments(state, 1, params->chunk_page_prep);
	params->exit_write_threads = false;
	deinit_align_buffers(state);
}

/** Calculate the total length of current frame */
int64_t camogm_get_jpeg_size(camogm_state *state)
{
	int64_t len = 0;

	for (int i = 0; i < state->chunk_index - 1; i++) {
		len += state->packetchunks[i + 1].bytes;
	}

	return len;
}

/**
 * @brief Called every time the JPEG files recording is started.
 *
 * This function checks if the raw device write is initiated and tries to open the device specified. The device
 * will be closed in #camogm_end_jpeg function.
 * @param[in]   state   a pointer to a structure containing current state
 * @return      0 if the device was opened successfully and negative error code otherwise
 */
int camogm_start_jpeg(camogm_state *state)
{
	char * slash;
	int rslt; // , i;
//	uint64_t offset;
	struct writer_params *params = &state->writer_params;
	uint64_t current_lba = params->lba_start;
	params->circbuf_many = false;
	if (!state->rawdev_op) {
		strcpy(state->path, state->path_prefix);   // make state->path a directory name (will be replaced when the frames will be written)
		slash = strrchr(state->path, '/');
		D2(fprintf(debug_file, "camogm_start_jpeg\n"));
		if (slash) {
			D3(fprintf(debug_file, "Full path %s\n", state->path));
			slash[0] = '\0';                       // truncate path to the directory name
			D3(fprintf(debug_file, "directory path %s\n", state->path));
			rslt = mkdir(state->path, 0777);
			D3(fprintf(debug_file, "mkdir (%s, 0777) returned %d, errno=%d\n", state->path, rslt, errno));
			if ((rslt < 0) && (errno != EEXIST)) { // already exists is OK
				D0(fprintf(debug_file, "Error creating directory %s, errno=%d\n", state->path, errno));
				return -CAMOGM_FRAME_FILE_ERR;
			}
		}
	} else {
#if 0
		state->writer_params.chunk_page_prep =  0; // was in init. Or should it continue and rely thatall prepared should be cleared when closing file
		for (i = 0; i < SEGMENTS_PAGES; i++){
			state->writer_params.chunk_page_state[i] = SEGPAGE_EMPTY;
		}
		state->writer_params.writev_run =       false;
#endif
		D6(fprintf(debug_file, "start_jpeg() before open_state_file: Block device positions: start = %llu, current = %llu, end = %llu, params->next_segment_pos=0x%llx\n",
				state->writer_params.lba_start, current_lba, state->writer_params.lba_end, state->writer_params.next_segment_pos));

		if ((rslt=open_state_file(&state->rawdev, &current_lba)) != 0){ // }&params->lba_current) != 0) {
			D0(fprintf(debug_file, "Could not get write pointer from state file (err=%d), recording will start from the rounded to page beginning of partition: "
					"%s\n", rslt, state->rawdev.rawdev_path));
		}
///		params->next_segment_pos = lba_to_offset(params->lba_current - params->lba_start);

		if ((current_lba < state->writer_params.lba_start) || (current_lba > state->writer_params.lba_end)) {
			D0(fprintf(debug_file, "start_jpeg(): INVALID start lba: start = %llu, current = %llu, end = %llu, resetting current_lba to lba_start\n",
					state->writer_params.lba_start, current_lba, state->writer_params.lba_end)); // incorrect
			current_lba=state->writer_params.lba_start;
		}
		state->writer_params.next_segment_pos = lba_to_offset(current_lba - state->writer_params.lba_start);
		D3(fprintf(debug_file, "start_jpeg(): Block device positions: start = %llu, current = %llu, end = %llu, params->next_segment_pos=0x%llx\n",
				state->writer_params.lba_start, current_lba, state->writer_params.lba_end, state->writer_params.next_segment_pos)); // incorrect

		params->next_segment_pos = lba_to_offset(current_lba - params->lba_start);
		D3(syslog (LOG_INFO, "start_jpeg(): lba_start = %lld, lba_end=%lld, current_lba=%lld, state->path=%s",\
				 params->lba_start,  params->lba_end,  get_lba_next(params), state->path));

		D2(fprintf(debug_file, "First write to block device: %s, will start with offset in bytes: %llu\n", \
				state->rawdev.rawdev_path, params->next_segment_pos));
		for (int i = 0; i < NUM_WRITER_THREADS; i++){
			params->writer_threads[i] = 0;
		}
		params->exit_write_threads = false;
		for (int i = 0; i < NUM_WRITER_THREADS; i++){
			rslt = pthread_create(&params->writer_threads[i], NULL, jpeg_writer, (void *)state);
			if (rslt != 0) {
				D0(fprintf(debug_file, "Can not start writer threads[%d]: %s\n", i, strerror(rslt)));
				return -CAMOGM_FRAME_FILE_ERR;
			}
			D6(fprintf(debug_file, "Started writer thread [%d] id=%ld OK\n", i, params->writer_threads[i]));
		}
		params->stat_update = get_fpga_time64(state->fd_fparmsall[0], 0);
		D1(fprintf(debug_file, "Started %d writer threads\n", NUM_WRITER_THREADS));
		// check if immediately exited because of open error? if (params->exit_write_threads) - now and while waiting for free buffer ?
	}
	return 0;
}

/**
 * @brief Write single JPEG frame
 *
 * This function will write single JPEG file
 * @param   state   a pointer to a structure containing current state
 * @return
 */
int camogm_frame_jpeg(camogm_state *state)
{
	int i, j;
	int dbg_start_wait = 0, dbg_wait;
	ssize_t iovlen, l = 0;
	struct iovec chunks_iovec[8];
	int port = state->port_num;
//	time_t curr_time;
	uint64_t curr_time;
	int is_tiff = state->this_frame_params[port].color == COLORMODE_RAW;
	struct writer_params *params = &state->writer_params;
	int seg_page = params->chunk_page_prep;
	uint64_t disk_size = (params->lba_end - params->lba_start) * PHY_BLOCK_SIZE;
    if (!state->rawdev_op) {
        if (is_tiff) {
        	sprintf(state->path, "%s%d_%010ld_%06ld.tiff", state->path_prefix, port, state->this_frame_params[port].timestamp_sec, state->this_frame_params[port].timestamp_usec);
        } else {
        	sprintf(state->path, "%s%d_%010ld_%06ld.jpeg", state->path_prefix, port, state->this_frame_params[port].timestamp_sec, state->this_frame_params[port].timestamp_usec);
        }
    	l = 0;
    	for (i = 0; i < (state->chunk_index) - 1; i++) {
    		chunks_iovec[i].iov_base = state->packetchunks[i + 1].chunk;
    		chunks_iovec[i].iov_len = state->packetchunks[i + 1].bytes;
    		l += chunks_iovec[i].iov_len;
    	}
    	if (((state->ivf = open(state->path, O_RDWR | O_CREAT, 0777))) < 0) {
    		D0(fprintf(debug_file, "Error opening %s for writing, returned %d, errno=%d\n", state->path, state->ivf, errno));
    		return -CAMOGM_FRAME_FILE_ERR;
    	}
    	iovlen = emul_writev(state->ivf, chunks_iovec, (state->chunk_index) - 1); // FIXME! Remove gaps?
    	if (iovlen < l) {
    		j = errno;
    		D0(fprintf(debug_file, "emul_writev error %d (returned %d, expected %d)\n", j, iovlen, l));
    		close(state->ivf);
    		return -CAMOGM_FRAME_FILE_ERR;
    	}
    	state->rawdev.last_jpeg_size = l;
    	close(state->ivf);
    } else {

    	D6(fprintf(debug_file, "ccam_dma_buf[%d] = %p\n",port, get_ccam_dma_buf(port)));
    	D3(fprintf(debug_file, "\n"));
    	D4(fprintf(debug_file, "_12.7_ camogm_frame_jpeg(): %ld.%06ld:%d @ %07d\n",state->this_frame_params[port].timestamp_sec, state->this_frame_params[port].timestamp_usec, port, get_fpga_usec(state->fd_fparmsall[0], 0)));
    	D6(fprintf(debug_file, "\ndump iovect array for port %u, chunk_index=%u, chunk_data_index=%u, data_segments=%u\n", state->port_num, state->chunk_index, state->chunk_data_index, state->data_segments));
    	for (int i = 0; i < state->chunk_index; i++) {
    		D6(fprintf(debug_file, "ptr: %p, length: %ld\n", state->packetchunks[i].chunk, state->packetchunks[i].bytes));
    	}
    	// next frame is ready for recording, signal this to the writer thread
#if NUM_WRITER_THREADS > 1
    	D6(fprintf(debug_file, "_13_: Yielding thread before getting a lock (%d):[%d%d:%d%d] @ %07d\n",\
    			params->chunk_page_prep, params->writev_run[0], params->writev_run[1], \
				params->write_waits_sig[0], params->write_waits_sig[1], get_fpga_usec(state->fd_fparmsall[0], 0)));
#else
    	D6(fprintf(debug_file, "_13_: Yielding thread before getting a lock (%d):[%d :%d ] @ %07d\n",\
    			params->chunk_page_prep, params->writev_run[0], \
				params->write_waits_sig[0], get_fpga_usec(state->fd_fparmsall[0], 0)));
#endif
    	pthread_yield();
    	pthread_mutex_lock(&params->writer_mutex);
///    	params->circbuf_many = get_num_empty(params) >= NUM_NEED_NEMPTY;
        if (get_num_empty(params) < NUM_FREE_MANY) {
        	params->circbuf_many = true;
        }
#if NUM_WRITER_THREADS > 1
    	D3(dbg_start_wait =  get_fpga_usec(state->fd_fparmsall[0], 0); fprintf(debug_file, \
    			"_13a_: waiting for empty buffer (%d): [%d%d%d%d%d%d%d%d] LOCK [%d%d:%d%d:%d%d] @ %07d\n",\
    			params->chunk_page_prep, \
				params->chunk_page_state[0],params->chunk_page_state[1],params->chunk_page_state[2],params->chunk_page_state[3], \
				params->chunk_page_state[4],params->chunk_page_state[5],params->chunk_page_state[6],params->chunk_page_state[7], \
				params->writev_run[0], params->writev_run[1], \
				params->write_waits_sig[0], params->write_waits_sig[1], \
				params->write_go[0],        params->write_go[1], \
				dbg_start_wait));
#else
    	D3(dbg_start_wait =  get_fpga_usec(state->fd_fparmsall[0], 0); fprintf(debug_file, \
    			"_13a_: waiting for empty buffer (%d): [%d%d%d%d%d%d%d%d] LOCK [%d:%d:%d] @ %07d\n",\
    			params->chunk_page_prep, \
				params->chunk_page_state[0],params->chunk_page_state[1],params->chunk_page_state[2],params->chunk_page_state[3], \
				params->chunk_page_state[4],params->chunk_page_state[5],params->chunk_page_state[6],params->chunk_page_state[7], \
				params->writev_run[0], \
				params->write_waits_sig[0], \
				params->write_go[0], \
				dbg_start_wait));
#endif
    	// do not run ahead by >= FILE_CHUNKS_PAGES (0 or 1 for just FILE_CHUNKS_PAGES == 2)
    	// either buffer is completely empty (state->writer_params->chunk_page_prep == state->writer_params->chunk_page_write)
    	// or it is not full and writev_run is active
#if NUM_WRITER_THREADS > 1
  #if	THREADS_MODE == 0
    	while (((i = get_num_empty(params)) < NUM_NEED_EMPTY) && !params->exit_write_threads) { // get_num_empty(params)< 3?
    		D6(fprintf(debug_file, "_13b_: waiting for empty buffer (%d):[%d%d:%d%d] @ %07d\n",\
    				params->chunk_page_prep, params->writev_run[0], params->writev_run[1], \
					params->write_waits_sig[0], params->write_waits_sig[1], get_fpga_usec(state->fd_fparmsall[0], 0)));
    		pthread_cond_wait(&params->main_cond, &params->writer_mutex);
    	} // pthread_cond_signal, pthread_cond_broadcast - signal or broadcast a condition
  #elif THREADS_MODE > 0
    	while ((((i = get_num_empty(params)) < NUM_NEED_EMPTY)
//    			|| ()
    			) && !params->exit_write_threads) { // get_num_empty(params)< 3?
    		D6(fprintf(debug_file, "_13b_: waiting for empty buffer (%d):[%d%d:%d%d] @ %07d\n",\
    				params->chunk_page_prep, params->writev_run[0], params->writev_run[1], \
					params->write_waits_sig[0], params->write_waits_sig[1], get_fpga_usec(state->fd_fparmsall[0], 0)));
    		pthread_cond_wait(&params->main_cond, &params->writer_mutex);
    	} // pthread_cond_signal, pthread_cond_broadcast - signal or broadcast a condition
  #endif
#else // single write thread
    	while (((i = get_num_empty(params)) < NUM_NEED_EMPTY) && !params->exit_write_threads) { // get_num_empty(params)< 3?
    		D6(fprintf(debug_file, "_13b_: waiting for empty buffer (%d):[%d:%d] @ %07d\n",\
    				params->chunk_page_prep, params->writev_run[0], \
					params->write_waits_sig[0],  get_fpga_usec(state->fd_fparmsall[0], 0)));
    		pthread_cond_wait(&params->main_cond, &params->writer_mutex);
    	} // pthread_cond_signal, pthread_cond_broadcast - signal or broadcast a condition

#endif
    	pthread_mutex_unlock(&params->writer_mutex);
#if NUM_WRITER_THREADS > 1
    	D3(dbg_wait=get_fpga_usec(state->fd_fparmsall[0], 0) - dbg_start_wait; if (dbg_wait <0) dbg_wait+=10000000;
    			fprintf(debug_file, "_13b0_: got empty buffer at %d (%d pages): [%d%d%d%d%d%d%d%d] [%d%d:%d%d:%d%d] @ %07d (+%07d)\n",\
    			params->chunk_page_prep, i, \
				params->chunk_page_state[0],params->chunk_page_state[1],params->chunk_page_state[2],params->chunk_page_state[3], \
				params->chunk_page_state[4],params->chunk_page_state[5],params->chunk_page_state[6],params->chunk_page_state[7], \
				params->writev_run[0], params->writev_run[1], \
				params->write_waits_sig[0], params->write_waits_sig[1], \
				params->write_go[0],        params->write_go[1], \
				get_fpga_usec(state->fd_fparmsall[0], 0), dbg_wait));
#else
    	D3(dbg_wait=get_fpga_usec(state->fd_fparmsall[0], 0) - dbg_start_wait; if (dbg_wait <0) dbg_wait+=10000000;
    			fprintf(debug_file, "_13b0_: got empty buffer at %d (%d pages): [%d%d%d%d%d%d%d%d] [%d:%d:%d] @ %07d (+%07d)\n",\
    			params->chunk_page_prep, i, \
				params->chunk_page_state[0],params->chunk_page_state[1],params->chunk_page_state[2],params->chunk_page_state[3], \
				params->chunk_page_state[4],params->chunk_page_state[5],params->chunk_page_state[6],params->chunk_page_state[7], \
				params->writev_run[0], \
				params->write_waits_sig[0], \
				params->write_go[0], \
				get_fpga_usec(state->fd_fparmsall[0], 0), dbg_wait));
#endif

    	if (params->exit_write_threads) { // at least one of the threads abnormally exited
        	D0(fprintf(debug_file, "_13b1_:exit_write_threads is true, abnormally exiting threads @ %07d\n",get_fpga_usec(state->fd_fparmsall[0], 0)));
    		for (int i = 0; i < NUM_WRITER_THREADS; i++){
    			if (params->writer_threads[i]) {
    	        	D0(fprintf(debug_file, "_13b2_:exiting writer thread % d @ %07d\n", i, get_fpga_usec(state->fd_fparmsall[0], 0)));
    				pthread_join(params->writer_threads[i], NULL);
    				params->writer_threads[i] = 0;
    			}
    		}
        	D0(fprintf(debug_file, "_13b3_:all writer threads exited @ %07d\n",get_fpga_usec(state->fd_fparmsall[0], 0)));
        	return -CAMOGM_FRAME_FILE_ERR;
    	}
#if NUM_WRITER_THREADS > 1
    	D6(fprintf(debug_file, "_13c_: got empty buffer         (%d):[%d%d:%d%d] @ %07d\n",\
    			params->chunk_page_prep, params->writev_run[0], params->writev_run[1], \
				params->write_waits_sig[0], params->write_waits_sig[1], get_fpga_usec(state->fd_fparmsall[0], 0)));
#else
    	D6(fprintf(debug_file, "_13c_: got empty buffer         (%d):[%d:%d] @ %07d\n",\
    			params->chunk_page_prep, params->writev_run[0], \
				params->write_waits_sig[0], get_fpga_usec(state->fd_fparmsall[0], 0)));
#endif
//		params->dbg_data[params->chunk_page_prep][0] = port; // inside remap_vectors(state)
    	remap_vectors(state); // now can be used instead of align_frame(state), returns total length
    	// save segment file position, update next
    	params->segment_pos[seg_page] = params->next_segment_pos;
    	for (i = 0; i < SEGMENTS_NUMBER; i++){
    		params->next_segment_pos += params->data_segments[seg_page][i].iov_len;
    	}
    	// wrap over file size
    	if (params->next_segment_pos >= disk_size){
    		params->next_segment_pos -= disk_size;
    	}

    	// proceed if last frame was recorded without errors
    	D6(fprintf(debug_file, "_13d_:frame aligned @ %07d\n",get_fpga_usec(state->fd_fparmsall[0], 0)));
    	pthread_mutex_lock(&params->writer_mutex);
    	if (params->last_ret_val == 0) {
	    	params->chunk_page_state[params->chunk_page_prep] = SEGPAGE_FULL; // mark as full;
    		params->chunk_page_prep = (params->chunk_page_prep + 1) % SEGMENTS_PAGES; // ***************** Here NEXT SEGMENT

#if NUM_WRITER_THREADS > 1
    		if (params->write_waits_sig[0] || params->write_waits_sig[1]) {
    			pthread_cond_broadcast(&params->writer_cond);
    			D3(fprintf(debug_file, "_13e_: chunk_page_prep = %d, signaled writer_cond * [%d%d:%d%d:%d%d] @ %07d\n", \
    					params->chunk_page_prep, params->writev_run[0], params->writev_run[1], \
						params->write_waits_sig[0], params->write_waits_sig[1], \
						params->write_go[0],        params->write_go[1], \
						get_fpga_usec(state->fd_fparmsall[0], 0)));
    		} else {
    			D3(fprintf(debug_file, "_13e0_: chunk_page_prep = %d, nobody listens, no signal -[%d%d:%d%d:%d%d] @ %07d\n", \
    					params->chunk_page_prep, params->writev_run[0], params->writev_run[1], \
						params->write_waits_sig[0], params->write_waits_sig[1], \
						params->write_go[0],        params->write_go[1], \
						get_fpga_usec(state->fd_fparmsall[0], 0)));
    		}
#else
    		if (params->write_waits_sig[0]) {
    			pthread_cond_broadcast(&params->writer_cond);
    			D3(fprintf(debug_file, "_13es_: chunk_page_prep = %d, signaled writer_cond * [%d:%d:%d] @ %07d\n", \
    					params->chunk_page_prep, params->writev_run[0], \
						params->write_waits_sig[0], \
						params->write_go[0], \
						get_fpga_usec(state->fd_fparmsall[0], 0)));
    		} else {
    			D3(fprintf(debug_file, "_13e0s_: chunk_page_prep = %d, nobody listens, no signal -[%d:%d:%d] @ %07d\n", \
    					params->chunk_page_prep, params->writev_run[0], \
						params->write_waits_sig[0], \
						params->write_go[0], \
						get_fpga_usec(state->fd_fparmsall[0], 0)));
    		}
#endif
    	}
    	pthread_mutex_unlock(&params->writer_mutex);
// add yield here?
    	D4(fprintf(debug_file, "_13e2a_ @ %07d\n", get_fpga_usec(state->fd_fparmsall[0], 0)));
    	if (params->last_ret_val != 0) {
    		return params->last_ret_val;
    	}
    	D4(fprintf(debug_file, "_13e2b_ @ %07d\n", get_fpga_usec(state->fd_fparmsall[0], 0)));
    	// update status file if time has come
//    	curr_time = time(NULL);
    	curr_time = get_fpga_time64(state->fd_fparmsall[0], 0);
//    	if (difftime(curr_time, params->stat_update) > STAT_UPDATE_PERIOD) {
       	if ((curr_time - params->stat_update) > (STAT_UPDATE_PERIOD * 1000000)) { // STAT_UPDATE_PERIOD is in seconds
    		save_state_file(state);
    		params->stat_update = curr_time;
    		D3(fprintf(debug_file, "_13f_: saved state file @ %07d\n", get_fpga_usec(state->fd_fparmsall[0], 0)));
    	}
    }
	D4(fprintf(debug_file, "_13e2_ camogm_frame_jpeg-> @ %07d\n", get_fpga_usec(state->fd_fparmsall[0], 0)));
    return 0;
}

/**
 * @brief Finish JPEG file write operation
 *
 * This function checks whether raw device write was on and closes raw device file.
 * @param   state   a pointer to a structure containing current state
 * @return  0 if the device was closed successfully and -1 otherwise
 */

int camogm_end_jpeg(camogm_state *state)
{
	int i, ret = 0;
	int bytes, ceil_size;
	struct writer_params *params = &state->writer_params;
	int seg_page = params->chunk_page_prep;
	uint64_t disk_size = (params->lba_end - params->lba_start) * PHY_BLOCK_SIZE;
	if (state->rawdev_op) {
		bytes = params->glue_carry_vec.iov_len; // is not modified by writer threads
		if (bytes > 0) {
			ceil_size = vectaligntail(&params->glue_carry_vec);
			D1(fprintf(debug_file, "Write last block of data, size = %d, full size = %d\n", bytes, ceil_size));
			pthread_mutex_lock(&params->writer_mutex);
			params->circbuf_many = true; // finish slowly?
#if NUM_WRITER_THREADS > 1
			D3(fprintf(debug_file, \
					"_x13a_: waiting for empty buffer (%d): [%d%d%d%d%d%d%d%d] LOCK [%d%d:%d%d:%d%d] @ %07d\n",\
					params->chunk_page_prep, \
					params->chunk_page_state[0],params->chunk_page_state[1],params->chunk_page_state[2],params->chunk_page_state[3], \
					params->chunk_page_state[4],params->chunk_page_state[5],params->chunk_page_state[6],params->chunk_page_state[7], \
					params->writev_run[0], params->writev_run[1], \
					params->write_waits_sig[0], params->write_waits_sig[1], \
					params->write_go[0],        params->write_go[1], \
					get_fpga_usec(state->fd_fparmsall[0], 0)));
			// Wait for the segment buffer (same as for the next frame)
#if 0
			while ((params->chunk_page_state[params->chunk_page_prep] != SEGPAGE_EMPTY) || // next page is busy, can not go ahead
					// TODO: Update?
					(!params->writev_run[0] && !params->writev_run[1] && (get_write_page(params) >= 0))) {
				D3(fprintf(debug_file, "_x13b_: waiting for empty buffer (%d):[%d%d:%d%d] @ %07d\n", \
						params->chunk_page_prep, params->writev_run[0], params->writev_run[1], \
						params->write_waits_sig[0], params->write_waits_sig[1], get_fpga_usec(state->fd_fparmsall[0], 0)));
				pthread_cond_wait(&params->main_cond, &params->writer_mutex);
			}
#endif
#else
			D3(fprintf(debug_file, \
					"_x13a_: waiting for empty buffer (%d): [%d%d%d%d%d%d%d%d] LOCK [%d:%d:%d] @ %07d\n",\
					params->chunk_page_prep, \
					params->chunk_page_state[0],params->chunk_page_state[1],params->chunk_page_state[2],params->chunk_page_state[3], \
					params->chunk_page_state[4],params->chunk_page_state[5],params->chunk_page_state[6],params->chunk_page_state[7], \
					params->writev_run[0], \
					params->write_waits_sig[0], \
					params->write_go[0], \
					get_fpga_usec(state->fd_fparmsall[0], 0)));
#endif

#if NUM_WRITER_THREADS > 1
  #if	THREADS_MODE == 0
			while (((i = get_num_empty(params)) < NUM_NEED_EMPTY) && !params->exit_write_threads) { // get_num_empty(params)< 3?
				D3(fprintf(debug_file, "_x13b_: waiting for empty buffer (%d):[%d%d:%d%d] @ %07d\n",\
						params->chunk_page_prep, params->writev_run[0], params->writev_run[1], \
						params->write_waits_sig[0], params->write_waits_sig[1], get_fpga_usec(state->fd_fparmsall[0], 0)));
				pthread_cond_wait(&params->main_cond, &params->writer_mutex);
			} // pthread_cond_signal, pthread_cond_broadcast - signal or broadcast a condition
  #elif THREADS_MODE > 0
			while ((((i = get_num_empty(params)) < NUM_NEED_EMPTY)
					) && !params->exit_write_threads) { // get_num_empty(params)< 3?
				D3(fprintf(debug_file, "_x13b_: waiting for empty buffer (%d):[%d%d:%d%d] @ %07d\n",\
						params->chunk_page_prep, params->writev_run[0], params->writev_run[1], \
						params->write_waits_sig[0], params->write_waits_sig[1], get_fpga_usec(state->fd_fparmsall[0], 0)));
				pthread_cond_wait(&params->main_cond, &params->writer_mutex);
			} // pthread_cond_signal, pthread_cond_broadcast - signal or broadcast a condition
  #endif
#else
			while (((i = get_num_empty(params)) < NUM_NEED_EMPTY) && !params->exit_write_threads) { // get_num_empty(params)< 3?
				D3(fprintf(debug_file, "_x13b_: waiting for empty buffer (%d):[%d:%d] @ %07d\n",\
						params->chunk_page_prep, params->writev_run[0], \
						params->write_waits_sig[0],  get_fpga_usec(state->fd_fparmsall[0], 0)));
				pthread_cond_wait(&params->main_cond, &params->writer_mutex);
			} // pthread_cond_signal, pthread_cond_broadcast - signal or broadcast a condition
#endif
			pthread_mutex_unlock(&params->writer_mutex);
			// prepare last segment, increment segment page
			params->data_segments[seg_page][SEGMENT_GLUE].iov_base =  params->glue_carry_vec.iov_base;
			params->data_segments[seg_page][SEGMENT_GLUE].iov_len =   params->glue_carry_vec.iov_len;
			params->data_segments[seg_page][SEGMENT_FIRST].iov_len =  0; // check iov_len first before using iov_base
			params->data_segments[seg_page][SEGMENT_SECOND].iov_len = 0; // check iov_len first before using iov_base
			params->segment_pos[seg_page] = params->next_segment_pos;
			params->next_segment_pos += params->data_segments[seg_page][SEGMENT_GLUE].iov_len;
			// wrap over file size
			if (params->next_segment_pos >= disk_size){
				params->next_segment_pos -= disk_size;
			}
			D3(fprintf(debug_file, "_x13d_:frame aligned @ %07d\n",get_fpga_usec(state->fd_fparmsall[0], 0)));
			pthread_mutex_lock(&params->writer_mutex);
			params->chunk_page_state[params->chunk_page_prep] = SEGPAGE_FULL;
			params->chunk_page_prep = (params->chunk_page_prep + 1) % SEGMENTS_PAGES; // ***************** Here NEXT SEGMENT (last time)
			params->exit_write_threads = true; // finish job and exit
			pthread_cond_broadcast(&params->writer_cond);
			D3(fprintf(debug_file, "_x13e1_: chunk_page_prep = %d, signal writer_cond @ %07d\n",params->chunk_page_prep, get_fpga_usec(state->fd_fparmsall[0], 0)));
			pthread_mutex_unlock(&params->writer_mutex);
		} // if (bytes > 0) {
		D2(fprintf(debug_file, "_x13f0_: will run save_state_file @ %07d\n", get_fpga_usec(state->fd_fparmsall[0], 0)));
		save_state_file(state);
		params->stat_update = get_fpga_time64(state->fd_fparmsall[0], 0);
		D2(fprintf(debug_file, "_x13f_: saved state file after camogm_end_jpeg() @ %07d\n", get_fpga_usec(state->fd_fparmsall[0], 0)));
		// terminate writer threads (consider moving create/join threads into initialization/destruction
		pthread_mutex_lock(&params->writer_mutex);
		params->exit_write_threads = true;
		pthread_cond_broadcast(&params->writer_cond);
		pthread_mutex_unlock(&params->writer_mutex);
		for (int i = 0; i < NUM_WRITER_THREADS; i++){
			if (params->writer_threads[i]) {
				pthread_join(params->writer_threads[i], NULL);
			}
		}
		D2(fprintf(debug_file, "_x13g_: joined writer threads @ %07d\n", get_fpga_usec(state->fd_fparmsall[0], 0)));

	} // (state->rawdev_op) {
	return ret;
}


#define NUM_TRIES 10
/**
 * @brief Disk writing thread. This thread holds local copy of a structure containing current state
 * of the program and updates it on a signal from main thread every time new frame is ready for recording.
 * @param[in]   thread_args   a pointer to a structure containing current state
 * @return      None
 */
void *jpeg_writer(void *thread_args)
{
	int dbg_start_wait = 0, dbg_wait, dbg_time;
	char * dbg_pref;
#define MAX_SEGM_NUMBER 16 // temporary, debugging
	struct iovec write_segments[MAX_SEGM_NUMBER]; // SEGMENTS_NUMBER];// 3
	int num_write_segments, seg_len;
	int dbg_us1=0, dbg_us2=0, dbg_us; // save usec for difference
	pthread_t this_thread = pthread_self();
	int this_thread_index=-1;
	long long dbg_pos;
	int reset_rem;
	long long dbg_pos64;
	int i, i0, i1, wpage;
	char * bufp;

//	int32_t res32;
//	int64_t res64;
//	int errno32, errno64;
	ssize_t iovlen, image_len_padded, ll, l0, l1, blocks_write; // , wpage;
	ssize_t writev_wlen;
	struct iovec split_chunk;
	int blockdev_fd;   ///< file descriptor for open block device where frame will be recorded, local for the thread
	camogm_state *state = (camogm_state *)thread_args;
	struct writer_params *params = &state->writer_params;
	// find this thread index. Probably no retries are needed
	for (i0 = 0; i0 < NUM_TRIES; i0++){
		for (i = 0; i < NUM_WRITER_THREADS; i++){
			if (pthread_equal(params->writer_threads[i],this_thread)) {
//			if (params->writer_threads[i] == this_thread) {
				this_thread_index = i;
				break;
			}
		}
		if (this_thread_index >=0) {
			break;
		}
		pthread_yield();
	}
	if (this_thread_index < 0){ // should not happen
		D0(syslog (LOG_INFO, "ERROR: could not get thread index in %d tries: thread ID=%ld @ %07d", \
				NUM_TRIES, pthread_self(), get_fpga_usec(state->fd_fparmsall[0], 0)));
		D0(fprintf(debug_file, "ERROR: could not get thread index in %d tries: thread ID=%ld @ %07d\n", \
				NUM_TRIES, pthread_self(), get_fpga_usec(state->fd_fparmsall[0], 0)));
		this_thread_index=0;
	}
	params->write_go[this_thread_index] = false;

	dbg_pref = this_thread_index ? "    ":"  ";
	D2( fprintf(debug_file, "%sBefore open [%d%d%d%d%d%d%d%d], page_prep=%d, thread ID=%ld, thread index=%d\n", dbg_pref,\
			params->chunk_page_state[0],params->chunk_page_state[1],params->chunk_page_state[2],params->chunk_page_state[3], \
			params->chunk_page_state[4],params->chunk_page_state[5],params->chunk_page_state[6],params->chunk_page_state[7], \
			params->chunk_page_prep, pthread_self(), this_thread_index));
	blockdev_fd = open(state->rawdev.rawdev_path, O_RDWR | O_DIRECT);

	D1(syslog (LOG_INFO, "%sRaw %s is opened in 0x%08x mode, thread ID=%ld @ %07d", dbg_pref,\
			state->rawdev.rawdev_path, fcntl(blockdev_fd, F_GETFL), pthread_self(), get_fpga_usec(state->fd_fparmsall[0], 0)));
	D1(fprintf(debug_file, "%sRaw %s is opened in 0x%08x mode, thread ID=%ld @ %07d\n",  dbg_pref,\
			state->rawdev.rawdev_path, fcntl(blockdev_fd, F_GETFL), pthread_self(), get_fpga_usec(state->fd_fparmsall[0], 0)));
	pthread_mutex_lock(&params->writer_mutex);
	D2( fprintf(debug_file, "%s[%d %d %d %d %d %d %d %d], page_prep=%d, thread ID=%ld, lba_end=%lld\n",  dbg_pref,\
			params->chunk_page_state[0],params->chunk_page_state[1],params->chunk_page_state[2],params->chunk_page_state[3], \
			params->chunk_page_state[4],params->chunk_page_state[5],params->chunk_page_state[6],params->chunk_page_state[7], \
			params->chunk_page_prep, pthread_self(), params->lba_end));
	pthread_mutex_unlock(&params->writer_mutex);

	//pthread_self()
	if (blockdev_fd < 0) {
		D0(fprintf(debug_file, "%sError opening block device: %s, exiting thread\n", dbg_pref, state->rawdev.rawdev_path));
		pthread_cond_signal(&params->main_cond); // wake up main thread
		params->exit_write_threads = true;
		return NULL; //
	}

//strerror(errno),
	params->state = STATE_RUNNING; // TODO: Check meaning with multiple write threads
	pthread_yield();


	D2(fprintf(debug_file, "%s_w000_:  This thread index =%d (got in i0=%d) Before jpeg_writer loop @ %07d\n", dbg_pref,\
			this_thread_index,i0, get_fpga_usec(state->fd_fparmsall[0], 0))); // before the debug file is set !
	while (true) { // will break if no data and params->exit_write_threads
		pthread_mutex_lock(&params->writer_mutex);  // first time before the debug file is set !
		params->write_go[this_thread_index] = false;
		params->write_waits_sig[this_thread_index] = true;
#if NUM_WRITER_THREADS > 1
		D3(dbg_start_wait =  get_fpga_usec(state->fd_fparmsall[0], 0); \
		fprintf(debug_file, "%s_w00_: (%d) thread ID=%ld  got LOCK [%d%d] @ %07d\n", dbg_pref,\
				params->chunk_page_prep, pthread_self(),
				params->write_waits_sig[0], params->write_waits_sig[1], dbg_start_wait));
#else
		D3(dbg_start_wait =  get_fpga_usec(state->fd_fparmsall[0], 0); \
		fprintf(debug_file, "%s_w00_: (%d) thread ID=%ld  got LOCK [%d] @ %07d\n", dbg_pref,\
				params->chunk_page_prep, pthread_self(),
				params->write_waits_sig[0], dbg_start_wait));
#endif
		// will not start if there are less than 3 (num threads+1) ready pages so the latest one that have DMA open
		// will not be touched (its header can return zeros)
		// finishing off the remaining pages (slow) is done when no other write threads may be still running DMA
#if NUM_WRITER_THREADS > 1
  #if	THREADS_MODE == 0
		while ((((wpage = get_write_page(params)) < 0) ||
				(params->circbuf_many ?
						((i = get_num_empty(params)) >= (SEGMENTS_PAGES - NUM_WRITER_THREADS)) : // FULL+BYSY >= (threads+1)
						((i = get_num_busy(params)) > 0) ) // one-by-one
						|| (!params->write_waits_sig[1 - this_thread_index] && !params->write_go[this_thread_index]) // other write thread is not waiting, not already set go
		) && !params->exit_write_threads) {
  #if 1
			if (get_num_empty(params) >= (NUM_NEED_EMPTY + 2)){
				pthread_cond_signal(&params->main_cond);  // signal to main thread that prepares data
				D3(fprintf(debug_file, "%s_w00.2_: signaling main while waiting to start write [%d%d:%d%d:%d%d] @ %07d\n", \
						dbg_pref, \
						params->writev_run[0],      params->writev_run[1], \
						params->write_waits_sig[0], params->write_waits_sig[1], \
						params->write_go[0],        params->write_go[1], \
						get_fpga_usec(state->fd_fparmsall[0], 0)));
			}
  #endif
			pthread_cond_wait(&params->writer_cond, &params->writer_mutex); // They are called with mutex locked by the calling thread or undefined behaviour will result.
		}
  #elif THREADS_MODE > 0
		while ((((wpage = get_write_page(params)) < 0) ||
				(params->circbuf_many ?
						((i = get_num_empty(params)) >= (SEGMENTS_PAGES - NUM_WRITER_THREADS)) : // FULL+BYSY >= (threads+1)
						((i = get_num_busy(params)) > 0) ) // one-by-one
//						|| (!params->write_waits_sig[1 - this_thread_index] && !params->write_go[this_thread_index]) // other write thread is not waiting, not already set go
		) && !params->exit_write_threads) {
			// Wake up main thread while waiting
  #if 1
			if (get_num_empty(params) >= (NUM_NEED_EMPTY + 2)){
				pthread_cond_signal(&params->main_cond);  // signal to main thread that prepares data
				D3(fprintf(debug_file, "%s_w00.2_: signaling main while waiting to start write [%d%d:%d%d:%d%d] @ %07d\n", \
						dbg_pref, \
						params->writev_run[0],      params->writev_run[1], \
						params->write_waits_sig[0], params->write_waits_sig[1], \
						params->write_go[0],        params->write_go[1], \
						get_fpga_usec(state->fd_fparmsall[0], 0)));
			}
  #endif
			pthread_cond_wait(&params->writer_cond, &params->writer_mutex); // They are called with mutex locked by the calling thread or undefined behaviour will result.
		}
  #endif
#else // here NUM_WRITER_THREADS == 1
		while ((((wpage = get_write_page(params)) < 0) ||
				(params->circbuf_many ?
						((i = get_num_empty(params)) >= (SEGMENTS_PAGES - NUM_WRITER_THREADS)) : // FULL+BYSY >= (threads+1)
						((i = get_num_busy(params)) > 0) ) // one-by-one
		) && !params->exit_write_threads) {
  #if 1
			if (get_num_empty(params) >= (NUM_NEED_EMPTY + 2)){
				pthread_cond_signal(&params->main_cond);  // signal to main thread that prepares data
				D3(fprintf(debug_file, "_w00.2_: signaling main while waiting to start write [%d:%d:%d] @ %07d\n", \
						params->writev_run[0], \
						params->write_waits_sig[0], \
						params->write_go[0], \
						get_fpga_usec(state->fd_fparmsall[0], 0)));
			} else {
				D3(fprintf(debug_file, "w00.2x_: wpage=%d, circbuf_many=%d get_num_empty()=%d, get_num_busy()=%d  @ %07d\n", \
						wpage,
						params->circbuf_many,
						get_num_empty(params),
						get_num_busy(params),
						get_fpga_usec(state->fd_fparmsall[0], 0)
						));

			}
  #endif
			pthread_cond_wait(&params->writer_cond, &params->writer_mutex); // They are called with mutex locked by the calling thread or undefined behaviour will result.
		}


#endif

#if NUM_WRITER_THREADS > 1
		D3(dbg_time= get_fpga_usec(state->fd_fparmsall[0], 0); dbg_wait= dbg_time - dbg_start_wait; if (dbg_wait <0) dbg_wait+=10000000; \
		fprintf(debug_file, "%s_w00.3_: (%d, %d), empty/busy:%d many=%d thread ID=%ld got S+D [%d%d:%d%d:%d%d] @ %07d (+%07d)\n",  dbg_pref,\
				params->chunk_page_prep, wpage, i, params->circbuf_many, pthread_self(), \
				params->writev_run[0],      params->writev_run[1], \
				params->write_waits_sig[0], params->write_waits_sig[1], \
				params->write_go[0],        params->write_go[1], \
				dbg_time, dbg_wait)); // dbg_time
#else
		D3(dbg_time= get_fpga_usec(state->fd_fparmsall[0], 0); dbg_wait= dbg_time - dbg_start_wait; if (dbg_wait <0) dbg_wait+=10000000; \
		fprintf(debug_file, "%s_w00.3_: (%d, %d), empty/busy:%d many=%d thread ID=%ld got S+D [%d:%d:%d] @ %07d (+%07d)\n",  dbg_pref,\
				params->chunk_page_prep, wpage, i, params->circbuf_many, pthread_self(), \
				params->writev_run[0], \
				params->write_waits_sig[0], \
				params->write_go[0], \
				dbg_time, dbg_wait)); // dbg_time
#endif

#if NUM_WRITER_THREADS > 1 // nothong to do for a single thread
  #if	THREADS_MODE == 0
		if (params->write_waits_sig[1 - this_thread_index]) { // still waiting?
			params->write_go[1 - this_thread_index] = true;
			pthread_cond_broadcast(&params->writer_cond); // signal to other writer thread  waiting for this one to finish writing
		}
  #elif THREADS_MODE > 0
		if (params->write_waits_sig[1 - this_thread_index]) { // still waiting?
			params->write_go[1 - this_thread_index] = true;
			pthread_cond_broadcast(&params->writer_cond); // signal to other writer thread  waiting for this one to finish writing
		}
  #endif
#endif
		params->write_go[this_thread_index] = false;
		params->write_waits_sig[this_thread_index] = false;
		if (wpage>=0) {
			params->chunk_page_state[wpage] = SEGPAGE_BUSY;
		}
		pthread_mutex_unlock(&params->writer_mutex);
		// does it switch to main right here after unlock?
		D3(fprintf(debug_file, "%s_w00.5_: (%d, %d) thread ID=%ld @ %07d\n",  dbg_pref,\
				params->chunk_page_prep, wpage, pthread_self(), get_fpga_usec(state->fd_fparmsall[0], 0)));
		if (wpage >= 0) {
			D3(fprintf(debug_file, "%s_w01_: got write segment page %d (preparing %d) [%d%d%d%d%d%d%d%d] @ %07d\n", dbg_pref,\
					wpage, params->chunk_page_prep,
					params->chunk_page_state[0],params->chunk_page_state[1],params->chunk_page_state[2],params->chunk_page_state[3], \
					params->chunk_page_state[4],params->chunk_page_state[5],params->chunk_page_state[6],params->chunk_page_state[7], \
					get_fpga_usec(state->fd_fparmsall[0], 0)));




			image_len_padded = 0;
			reset_rem = 0;
			params->last_ret_val = 0;
			// inputs from params->data_chunks, returns number of result chunks
#define MAX_SEGM_PAGES  0xa0 // for large JPEG - there is some me introduced bug in the kernel that limits chunk size. maybe 0xff - OK?
#define MAX_SEGM_LENGTH	(MAX_SEGM_PAGES * PAGE_PHYS) // 0xa0000 // multiple of
			// Seems that I introduced a bug into kernel - it can not handle too long segments

			num_write_segments = 0;
			for (i = 0; i < SEGMENTS_NUMBER; i++){
				seg_len = params->data_segments[wpage][i].iov_len;
				image_len_padded+=seg_len;
				if (seg_len > 0){
					if (seg_len <= MAX_SEGM_LENGTH) {
						write_segments[num_write_segments++] = params->data_segments[wpage][i];
					} else {
						bufp = (char *) params->data_segments[wpage][i].iov_base;
						while (seg_len >= MAX_SEGM_LENGTH){
							write_segments[num_write_segments  ].iov_base = bufp;
							write_segments[num_write_segments++].iov_len = MAX_SEGM_LENGTH;
							bufp += MAX_SEGM_LENGTH;
							seg_len -= MAX_SEGM_LENGTH;
						}
						if (seg_len > 0){
							write_segments[num_write_segments  ].iov_base = bufp;
							write_segments[num_write_segments++].iov_len =  seg_len;
						}
					}
				}
			}
			// l is correct here
			// chunk_index = get_data_buffers(state, chunks_iovec[wpage], FILE_CHUNKS_NUM, wpage);
			// assuming write is block-aligned
			if (num_write_segments > 0) {
				// Moved rollover here
				blocks_write = image_len_padded / PHY_BLOCK_SIZE; // here LBA, 512
				if (get_lba_next(params) + blocks_write <= params->lba_end) { // almost always - single write
					D2(dbg_us1=get_fpga_usec(state->fd_fparmsall[0], 0));
					D3(dbg_pos = lseek64(blockdev_fd, 0, SEEK_CUR);  if (dbg_pos < 0) fprintf( \
							debug_file,"%slseek64()->-1, errno=%d(%s)",  dbg_pref,errno, strerror(errno)));

					D6(fprintf(debug_file, "%s_w02_: starting emul_writev@ %07d\n",  dbg_pref, dbg_us1));
					params->writev_run[this_thread_index] = true;
					dbg_pos64=lseek64(blockdev_fd, params->segment_pos[wpage], SEEK_SET);
					if (dbg_pos64 < 0){
						D0(fprintf(debug_file, "%slseek64(%d, %lld, %d)->%lld, errno=%d(%s)\n",  dbg_pref,\
								blockdev_fd, params->segment_pos[wpage], SEEK_SET,  dbg_pos64, errno, strerror(errno)));
					}

					iovlen = 0;
					for (i = 0; i < num_write_segments; i++){
						params->writev_run_segm[this_thread_index] = i;
						if (write_segments[i].iov_len & (PAGE_PHYS -1)){
							D0(fprintf(debug_file, "Length not page-aligned: while trying direct writing of %d bytes from %p, data = 0x%llx, segment=%d\n",\
									write_segments[i].iov_len, write_segments[i].iov_base, ((uint64_t*) write_segments[i].iov_base)[0], i ));
							D0(syslog (LOG_INFO, "Length not page-aligned: while trying direct writing of %d bytes from %p, data = 0x%llx, segment=%d",\
									write_segments[i].iov_len, write_segments[i].iov_base, ((uint64_t*) write_segments[i].iov_base)[0], i ));
							debug_level = 6;
							iovlen = -1;
							break;
						}

						if (write_segments[i].iov_len > 0) {
							if (i == 1) { // only for segment 1, if other write thread also is writing second (large) segment
								pthread_mutex_lock(&params->writer_mutex); // maybe it will not be needed - mutex

#if NUM_WRITER_THREADS > 1 // nothing to do here for a single thread
  #if	THREADS_MODE == 0
								if (params->writev_run[1 - this_thread_index] &&
										(params->writev_run_segm[1-this_thread_index] > 0) &&
										(get_num_empty(params) >= NUM_NEED_EMPTY)){
									pthread_cond_signal(&params->main_cond);  // signal to main thread that prepares data
									D3(dbg_us=get_fpga_usec(state->fd_fparmsall[0], 0);fprintf(debug_file, \
											"%s_w02a_: starting second (large) write and signaling main <- @ %07d (+%d)\n", \
											dbg_pref, dbg_us, ( dbg_us - dbg_us1)));
								}
  #elif THREADS_MODE == 1
								if (((params->writev_run[1 - this_thread_index] && // other is writing big chunk
										(params->writev_run_segm[1-this_thread_index] > 0))
										||(params->write_waits_sig[1 - this_thread_index]) // or it is waiting
										) &&
										(get_num_empty(params) >= NUM_NEED_EMPTY)){
									pthread_cond_signal(&params->main_cond);  // signal to main thread that prepares data
									D3(dbg_us=get_fpga_usec(state->fd_fparmsall[0], 0);fprintf(debug_file, \
											"%s_w02a_: starting second (large) write and signaling main <- @ %07d (+%d)\n", \
											dbg_pref, dbg_us, ( dbg_us - dbg_us1)));
								}
  #elif THREADS_MODE > 0
								// do nothing. Or do if it really wants?
  #endif
#else // single write thread
								if (get_num_empty(params) >= NUM_NEED_EMPTY){
									pthread_cond_signal(&params->main_cond);  // signal to main thread that prepares data
									D3(dbg_us=get_fpga_usec(state->fd_fparmsall[0], 0);fprintf(debug_file, \
											"%s_w02a_: starting second (large) write and signaling main <- @ %07d (+%d)\n", \
											dbg_pref, dbg_us, ( dbg_us - dbg_us1)));
								}
#endif
								pthread_mutex_unlock(&params->writer_mutex);
							}

							writev_wlen = write (blockdev_fd, write_segments[i].iov_base, write_segments[i].iov_len);
//							if (i == 1) { // only for segment 1, if other write thread also is writing second (large) segment
								D3(fprintf(debug_file, \
										"%s_w02x_: i=%d  @ %07d : start =0x%08x, len=0x%08x\n",\
										dbg_pref, i, get_fpga_usec(state->fd_fparmsall[0], 0), (int)write_segments[i].iov_base, (int)write_segments[i].iov_len));
//							}
							if (writev_wlen < 0) {
								D0(fprintf(debug_file, "Error %d while trying direct writing of %d bytes from %p, data = 0x%llx, segment=%d\n",\
										writev_wlen, write_segments[i].iov_len, write_segments[i].iov_base, ((uint64_t*) write_segments[i].iov_base)[0],i ));
								iovlen = writev_wlen;
								break;
							} else if (writev_wlen < write_segments[i].iov_len) {
								iovlen += writev_wlen;
								break;
							} else {
								iovlen += writev_wlen;
							}
						}
					}

					params->writev_run[this_thread_index] = false;
					D2(dbg_us=get_fpga_usec(state->fd_fparmsall[0], 0);	fprintf( \
							debug_file, "%s_w03_: finished emul_writev, wpage=%x, thread ID=%ld, pos= %lld  @ %07d (+ %07d, +%07d)\n",  dbg_pref,\
							wpage, pthread_self(), dbg_pos64, dbg_us, (dbg_us-dbg_us1), (dbg_us-dbg_us2)));
					D4(syslog (LOG_INFO, "%s_w03_: finished emul_writev, pos= %lld, len = %d(0x%x) thread ID = %ld @ %07d (+ %07d, +%07d)", dbg_pref,\
							dbg_pos, iovlen, iovlen, pthread_self(),  dbg_us, (dbg_us-dbg_us1), (dbg_us-dbg_us2)));
					D2(dbg_us2=dbg_us);
					if (iovlen < image_len_padded) {
						D0(fprintf(debug_file, "%semul_writev error: %s (returned %i, expected %i)\n", dbg_pref, strerror(errno), iovlen, image_len_padded));
						params->last_ret_val = -CAMOGM_FRAME_FILE_ERR;
						reset_rem = 1;
					} else {
						D6(fprintf(debug_file, "%sCurrent position in block device: %lld\n", dbg_pref, lseek64(blockdev_fd, 0, SEEK_CUR)));
					}
				} else { // Split into two writes: to the end of partition and startion from the beginning
					l0 = get_lba_next(params) * PHY_BLOCK_SIZE; // remaining to be written before rollover
					l1 = image_len_padded - l0; // remaining to be written after rollover
					i1 = 0; // start of the second write
					if (l0 > 0) { // should be always
						ll = 0;
						for (i0 = 0; i0 < num_write_segments; i0++){
							if (ll + write_segments[i0].iov_len > l0){
								break;
							}
							ll += write_segments[i0].iov_len;
						}
						// save split chunk
						split_chunk.iov_base = write_segments[i0].iov_base;
						split_chunk.iov_len =  write_segments[i0].iov_len;
						i1 = i0; // index of the first chunk in a second write
						if (ll < l0) { // is it a partial chunk?
							write_segments[i0].iov_len = l0 - ll;
							i0++;     // include split chunk in the first write
						}
						// write the first group (before rollover)
						params->writev_run[this_thread_index] = true;
						pthread_cond_signal(&params->main_cond);

						lseek64(blockdev_fd, params->segment_pos[wpage], SEEK_SET);
						iovlen = emul_writev(blockdev_fd, write_segments, i0);
						params->writev_run[this_thread_index] = false;
						lseek(blockdev_fd, 0, SEEK_SET); // regardless of error - rewind
						if (iovlen < l0) {
							D0(fprintf(debug_file, "%s(1 of 2) emul_writev error: %s (returned %i, expected %i)\n", dbg_pref, strerror(errno), iovlen, l0));
							params->last_ret_val = -CAMOGM_FRAME_FILE_ERR;
							reset_rem = 1;
						} else {
							// do not update statistic after the first emul_writev(), do it after the second
							D6(fprintf(debug_file, "%s(1 of 2) Current position in block device: %lld\n", dbg_pref, lseek64(blockdev_fd, 0, SEEK_CUR)));
							// update split chunk to use its second part, shift chunks
							if (l1 > 0) { // length to be written after rollover
								if (write_segments[i1].iov_len != split_chunk.iov_len) { // it was split
									D0(fprintf(debug_file, "%schunk %d was split, full length = %d, writtren = %d\n", dbg_pref, i1, split_chunk.iov_len, write_segments[i1].iov_len));
									write_segments[i1].iov_base += write_segments[i1].iov_len;
									write_segments[i1].iov_len = split_chunk.iov_len - write_segments[i1].iov_len;
								}
								// shift chunks
								for (i0 = i1; i0 < num_write_segments; i0++){
									write_segments[i0 - i1].iov_base = write_segments[i0].iov_base;
									write_segments[i0 - i1].iov_len =  write_segments[i0].iov_len;
								}
								params->writev_run[this_thread_index] = true;
								pthread_cond_signal(&params->main_cond);
								lseek64(blockdev_fd, 0, SEEK_SET);
								iovlen = emul_writev(blockdev_fd, write_segments, num_write_segments - i1);
								params->writev_run[this_thread_index] = false;
								if (iovlen < l1) {
									D0(fprintf(debug_file, "%s(2 of 2) emul_writev error: %s (returned %i, expected %i)\n", dbg_pref, strerror(errno), iovlen, l1));
									params->last_ret_val = -CAMOGM_FRAME_FILE_ERR;
									reset_rem = 1;
								} else {
									D6(fprintf(debug_file, "%s(2 of 2) Current position in block device: %lld\n", dbg_pref, lseek64(blockdev_fd, 0, SEEK_CUR)));
								}
							}
						}
					}
				}
				D6(fprintf(debug_file, "%s_w04: wrote data @ %07d\n", dbg_pref, get_fpga_usec(state->fd_fparmsall[0], 0)));
				D6(fprintf(debug_file, "%s_w05_: Block device positions: start = %llu, current = %llu, end = %llu, params->next_segment_pos=0x%llx\n", dbg_pref,\
						params->lba_start, get_lba_next(params), params->lba_end, params->next_segment_pos)); // incorrect
				if ((get_lba_next(params) > params->lba_end) || (get_lba_next(params) < params->lba_start)){
					D0(fprintf(debug_file, "%s_w05e_: ERROR: Block device positions: start = %llu, current = %llu, end = %llu, params->next_segment_pos=0x%llx\n", dbg_pref,\
							params->lba_start, get_lba_next(params), params->lba_end, params->next_segment_pos)); // incorrect
				}
			} else {
				D0(fprintf(debug_file, "%sdata vector mapping error: %d)\n",  dbg_pref, num_write_segments));
				params->last_ret_val = -CAMOGM_FRAME_FILE_ERR;
				reset_rem = 1;
			}
			// release main thread
			reset_segments(state, reset_rem, wpage);
			pthread_mutex_lock(&params->writer_mutex);
#if NUM_WRITER_THREADS > 1 // nothing to do here for a single thread
  #if THREADS_MODE == 2
			// signal main when one just ended all writes, and the other is either writing big or waiting

			if (((params->writev_run[1 - this_thread_index] && (params->writev_run_segm[1-this_thread_index] > 0))// other is writing big chunk
					||(params->write_waits_sig[1 - this_thread_index]) // or it is waiting
					) &&
					(get_num_empty(params) >= NUM_NEED_EMPTY)){
				pthread_cond_signal(&params->main_cond);  // signal to main thread that prepares data
				D3(dbg_us=get_fpga_usec(state->fd_fparmsall[0], 0);fprintf(debug_file, \
						"%s_w031_: finished, other is writing second (large), signaling main <- @ %07d (+%d)\n", \
						dbg_pref, dbg_us, ( dbg_us - dbg_us1)));
			}
  #endif
#endif
			params->chunk_page_state[wpage] = SEGPAGE_EMPTY;
// update statistics right here !
		    state->rawdev.last_jpeg_size = image_len_padded;
		    state->rawdev.total_rec_len += state->rawdev.last_jpeg_size;
		    D3(fprintf(debug_file, "%sl=%d, total_rec_len=%lld  [%d%d%d%d%d%d%d%d]\n",dbg_pref, image_len_padded, state->rawdev.total_rec_len,
					params->chunk_page_state[0],params->chunk_page_state[1],params->chunk_page_state[2],params->chunk_page_state[3], \
					params->chunk_page_state[4],params->chunk_page_state[5],params->chunk_page_state[6],params->chunk_page_state[7]));
//		    D6(fprintf(debug_file, "%sCurrent position in block device: %lld\n", dbg_pref, lseek64(blockdev_fd, 0, SEEK_CUR)));

// moved to while waiting
//			pthread_cond_signal(&params->main_cond); // signal to main thread that prepares data
			pthread_cond_broadcast(&params->writer_cond); // signal to other writer thread  waiting for this one to finish writing
			D6(fprintf(debug_file, "%s_w06b_: wrote data (%d , %d), signal main thread @ %07d\n", dbg_pref, \
					params->chunk_page_prep, wpage, get_fpga_usec(state->fd_fparmsall[0], 0)));

			pthread_mutex_unlock(&params->writer_mutex); // do not release untill waiting for emul_writev() or waiting for data to write
		} else {
			if (params->exit_write_threads) {
				break;
			}
		}
	}
	params->state = STATE_STOPPED;
	if (close(blockdev_fd) <0) {
		D0(fprintf(debug_file, "Error: %s\n", strerror(errno)));
	}
	// close raw file (one per thread) pthread_self(), get_fpga_usec(state->fd_fparmsall[0], 0)
	D1(fprintf(debug_file, "%s_w07_: Exit from recording thread ID=%ld, @ %07d\n",  dbg_pref, pthread_self(), get_fpga_usec(state->fd_fparmsall[0], 0)));
	D1(syslog (LOG_INFO,   "%sExit from recording thread ID=%ld, @ %07d", dbg_pref,\
				pthread_self(), get_fpga_usec(state->fd_fparmsall[0], 0)));
	return NULL;
}
